import os
import pathlib

from bq_schema.migration.table_finder import find_tables


def test_table_finder():
    file_path = pathlib.Path(__file__).parent
    tables_dir = os.path.join(file_path, "tables")
    table_names = {t.name for t in find_tables(tables_dir)}
    expected_table_names = {"first_table", "second_table"}
    assert expected_table_names == table_names



import json
import os

from google.cloud import bigquery
from google.cloud.bigquery import Client
from google.oauth2 import service_account


def create_connection() -> Client:
    service_file_json = os.environ.get("GOOGLE_SERVICE_FILE")
    if service_file_json:
        service_file = json.loads(service_file_json)
        credentials = service_account.Credentials.from_service_account_info(
            service_file
        )
        return Client(credentials=credentials, project=service_file["project_id"])

    return bigquery.Client()



from dataclasses import is_dataclass
from typing import List, Optional, Type, Union, cast

from google.cloud.bigquery import SchemaField, TimePartitioning

from bq_schema.dataclass_converter import dataclass_to_schema


# pylint: disable=missing-function-docstring
class BigqueryTable:
    """
    Configure a bigquery table.
    """

    @property
    def name(self) -> str:
        raise NotImplementedError

    @property
    def schema(self) -> Union[List[SchemaField], Type]:
        raise NotImplementedError

    @property
    def project(self) -> Optional[str]:
        return None

    @property
    def dataset(self) -> Optional[str]:
        return None

    @property
    def version(self) -> Optional[str]:
        return None

    @property
    def time_partitioning(self) -> Optional[TimePartitioning]:
        return None

    def full_table_name(self) -> str:
        """
        Generate the final table name.
        """
        if self.version:
            return f"{self.name}_v{self.version}"

        return self.name

    def get_schema_fields(self) -> List[SchemaField]:
        """
        Return the schema as a list of SchemaField.
        """
        if is_dataclass(self.schema):
            return dataclass_to_schema(cast(Type, self.schema))

        return cast(List[SchemaField], self.schema)



import inspect
from dataclasses import field
from datetime import date, datetime, time
from decimal import Decimal
from typing import List, Optional

from google.cloud.bigquery import SchemaField

from bq_schema.schema_converter import schema_to_dataclass
from bq_schema.types.type_mapping import Timestamp


# pylint: disable=line-too-long, too-few-public-methods
# fmt: off
class RequiredNestedField:
    int_field: int = field(metadata={"description": "This field is an INT field."})


class RequiredSchema:
    string_field: str = field(metadata={"description": "This field is a STRING field."})
    bytes_field: bytes
    int_field: int
    float_field: float
    numeric_field: Decimal
    bool_field: bool
    timestamp_field: Timestamp
    date_field: date
    time_field: time
    datetime_field: datetime
    required_nested_field: RequiredNestedField = field(metadata={"description": "This field is a STRUCT field."})
# fmt: on
# pylint: enable=line-too-long


def test_required_schema_to_dataclass():
    schema = [
        SchemaField(
            "string_field", "STRING", "REQUIRED", "This field is a STRING field.", ()
        ),
        SchemaField("bytes_field", "BYTES", "REQUIRED", None, ()),
        SchemaField("int_field", "INT64", "REQUIRED", None, ()),
        SchemaField("float_field", "FLOAT64", "REQUIRED", None, ()),
        SchemaField("numeric_field", "NUMERIC", "REQUIRED", None, ()),
        SchemaField("bool_field", "BOOL", "REQUIRED", None, ()),
        SchemaField("timestamp_field", "TIMESTAMP", "REQUIRED", None, ()),
        SchemaField("date_field", "DATE", "REQUIRED", None, ()),
        SchemaField("time_field", "TIME", "REQUIRED", None, ()),
        SchemaField("datetime_field", "DATETIME", "REQUIRED", None, ()),
        SchemaField(
            "required_nested_field",
            "STRUCT",
            "REQUIRED",
            "This field is a STRUCT field.",
            (
                SchemaField(
                    "int_field", "INT64", "REQUIRED", "This field is an INT field.", ()
                ),
            ),
        ),
    ]
    expected = (
        f"@dataclass\n{inspect.getsource(RequiredNestedField)}\n"
        f"@dataclass\n{inspect.getsource(RequiredSchema)}"
    )
    assert schema_to_dataclass("RequiredSchema", schema).strip() == expected.strip()


# pylint: disable=line-too-long
# fmt: off
class NullableNestedField:
    int_field: Optional[int] = field(metadata={"description": "This field is an INT field."})


class NullableSchema:
    string_field: Optional[str] = field(metadata={"description": "This field is a STRING field."})
    bytes_field: Optional[bytes]
    int_field: Optional[int]
    float_field: Optional[float]
    numeric_field: Optional[Decimal]
    bool_field: Optional[bool]
    timestamp_field: Optional[Timestamp]
    date_field: Optional[date]
    time_field: Optional[time]
    datetime_field: Optional[datetime]
    nullable_nested_field: Optional[NullableNestedField] = field(metadata={"description": "This field is a STRUCT field."})
# fmt: on
# pylint: enable=line-too-long


def test_optional_schema_to_dataclass():
    schema = [
        SchemaField(
            "string_field", "STRING", "NULLABLE", "This field is a STRING field.", ()
        ),
        SchemaField("bytes_field", "BYTES", "NULLABLE", None, ()),
        SchemaField("int_field", "INT64", "NULLABLE", None, ()),
        SchemaField("float_field", "FLOAT64", "NULLABLE", None, ()),
        SchemaField("numeric_field", "NUMERIC", "NULLABLE", None, ()),
        SchemaField("bool_field", "BOOL", "NULLABLE", None, ()),
        SchemaField("timestamp_field", "TIMESTAMP", "NULLABLE", None, ()),
        SchemaField("date_field", "DATE", "NULLABLE", None, ()),
        SchemaField("time_field", "TIME", "NULLABLE", None, ()),
        SchemaField("datetime_field", "DATETIME", "NULLABLE", None, ()),
        SchemaField(
            "nullable_nested_field",
            "STRUCT",
            "NULLABLE",
            "This field is a STRUCT field.",
            (
                SchemaField(
                    "int_field", "INT64", "NULLABLE", "This field is an INT field.", ()
                ),
            ),
        ),
    ]
    expected = (
        f"@dataclass\n{inspect.getsource(NullableNestedField)}\n"
        f"@dataclass\n{inspect.getsource(NullableSchema)}"
    )
    assert schema_to_dataclass("NullableSchema", schema) == expected.strip()


# pylint: disable=line-too-long
# fmt: off
class RepeatedNestedField:
    int_field: List[int] = field(metadata={"description": "This field is an INT field."})


class RepeatedSchema:
    string_field: List[str] = field(metadata={"description": "This field is a STRING field."})
    bytes_field: List[bytes]
    int_field: List[int]
    float_field: List[float]
    numeric_field: List[Decimal]
    bool_field: List[bool]
    timestamp_field: List[Timestamp]
    date_field: List[date]
    time_field: List[time]
    datetime_field: List[datetime]
    repeated_nested_field: List[RepeatedNestedField] = field(metadata={"description": "This field is a STRUCT field."})
# fmt: on
# pylint: enable=line-too-long


def test_repeated_schema_to_dataclass():
    schema = [
        SchemaField(
            "string_field", "STRING", "REPEATED", "This field is a STRING field.", ()
        ),
        SchemaField("bytes_field", "BYTES", "REPEATED", None, ()),
        SchemaField("int_field", "INT64", "REPEATED", None, ()),
        SchemaField("float_field", "FLOAT64", "REPEATED", None, ()),
        SchemaField("numeric_field", "NUMERIC", "REPEATED", None, ()),
        SchemaField("bool_field", "BOOL", "REPEATED", None, ()),
        SchemaField("timestamp_field", "TIMESTAMP", "REPEATED", None, ()),
        SchemaField("date_field", "DATE", "REPEATED", None, ()),
        SchemaField("time_field", "TIME", "REPEATED", None, ()),
        SchemaField("datetime_field", "DATETIME", "REPEATED", None, ()),
        SchemaField(
            "repeated_nested_field",
            "STRUCT",
            "REPEATED",
            "This field is a STRUCT field.",
            (
                SchemaField(
                    "int_field", "INT64", "REPEATED", "This field is an INT field.", ()
                ),
            ),
        ),
    ]
    expected = (
        f"@dataclass\n{inspect.getsource(RepeatedNestedField)}\n"
        f"@dataclass\n{inspect.getsource(RepeatedSchema)}"
    )
    assert schema_to_dataclass("RepeatedSchema", schema) == expected.strip()



import argparse
from argparse import Namespace
from pprint import pprint
from typing import Optional

from google.api_core.exceptions import NotFound
from google.cloud.bigquery.table import Table

from bq_schema.cli.bigquery_connection import create_connection
from bq_schema.migration.schema_diff import (
    apply_schema_differences,
    confirm_apply_schema_differences,
    find_schema_differences,
    print_format_schema_differences,
)
from bq_schema.migration.table_finder import find_tables


def parse_args() -> Namespace:
    parser = argparse.ArgumentParser(
        description="Manage your bigquery table migrations."
    )
    parser.add_argument("--project", required=False, help="Target bigquery project.")
    parser.add_argument("--dataset", required=False, help="Target bigquery dataset")
    parser.add_argument(
        "--module-path", required=True, help="Module path to your bigquery models."
    )

    parser.add_argument(
        "--apply",
        default=False,
        action="store_true",
        help="If set to true, this script will apply all open migrations.",
    )

    parser.add_argument(
        "--validate",
        default=False,
        action="store_true",
        help="If set to true, this script will fail if a difference in schemas is found. Useful for CI.",
    )

    return parser.parse_args()


def main(
    project: Optional[str],
    dataset: Optional[str],
    module_path: str,
    apply: bool,
    validate: bool,
) -> None:
    bigquery_client = create_connection()
    global_project = project
    global_dataset = dataset
    schemas_errors = {}

    print("Finding schema differences...")
    schema_diffs = find_schema_differences(
        module_path=module_path,
        bigquery_client=bigquery_client,
        global_project=project,
        global_dataset=dataset,
    )
    formated_schema_diff = print_format_schema_differences(schema_diffs=schema_diffs)
    if formated_schema_diff:
        print("Schema Differences:")
        pprint(formated_schema_diff)
        if validate:
            raise Exception(formated_schema_diff)
    else:
        print("No schema differences found.")

    if apply:
        if confirm_apply_schema_differences():
            apply_schema_differences(
                schema_diffs=schema_diffs, bigquery_client=bigquery_client
            )


def cli() -> None:
    args = parse_args()
    main(args.project, args.dataset, args.module_path, args.apply, args.validate)


if __name__ == "__main__":
    cli()



from invoke import task

_TEST_FOLDER = "tests"
_SOURCE_FOLDERS = " ".join(["bq_schema", _TEST_FOLDER])


@task
def lint(context):
    context.run(f"pylint {_SOURCE_FOLDERS}")


@task
def type_check(context):
    context.run("mypy bq_schema")


@task
def check_code_format(context):
    context.run("black --check .")
    context.run("isort --profile black --check .")


@task
def test(context):
    context.run(
        f"pytest {_TEST_FOLDER} --doctest-modules --junitxml=junit/test-results.xml --cov=bq_schema --cov-report=xml --cov-report=html"
    )


@task
def format_code(context):
    context.run("black .")
    context.run("isort --profile black .")


@task(pre=[lint, type_check, check_code_format, test])
def check_all(_):
    pass



from google.cloud.bigquery import SchemaField

from bq_schema.migration.schema_diff import check_schemas


def test_check_schemas():
    local_schema = [
        SchemaField("a", "RECORD", "REQUIRED", fields=[SchemaField("b", "INTEGER")])
    ]
    remote_schema = [
        SchemaField("a", "RECORD", "REQUIRED", fields=[SchemaField("b", "INTEGER")])
    ]
    assert list(check_schemas(local_schema, remote_schema)) == []


def test_find_new_columns():
    local_schema = [SchemaField("a", "INTEGER")]
    assert list(check_schemas(local_schema, [])) == [
        "New column SchemaField('a', 'INTEGER', 'NULLABLE', None, ())"
    ]


def test_find_new_columns_missing_record():
    local_schema = [SchemaField("a", "RECORD")]
    assert list(check_schemas(local_schema, [])) == [
        "New column SchemaField('a', 'RECORD', 'NULLABLE', None, ())"
    ]


def test_find_new_columns_missing_column_in_record():
    local_schema = [SchemaField("a", "RECORD", fields=[SchemaField("b", "INTEGER")])]
    remote_schema = [SchemaField("a", "RECORD", fields=[])]
    assert list(check_schemas(local_schema, remote_schema)) == [
        "Nested: New column SchemaField('b', 'INTEGER', 'NULLABLE', None, ())"
    ]


def test_check_mode_fail():
    local_schema = [
        SchemaField("a", "RECORD", "REQUIRED", fields=[SchemaField("b", "INTEGER")])
    ]
    remote_schema = [
        SchemaField("a", "RECORD", "NULLABLE", fields=[SchemaField("b", "INTEGER")])
    ]
    assert list(check_schemas(local_schema, remote_schema)) == [
        "There is difference between SchemaField('a', 'RECORD', 'REQUIRED', "
        "None, (SchemaField('b', 'INTEGER', 'NULLABLE', None, ()),)) and "
        "SchemaField('a', 'RECORD', 'NULLABLE', None, (SchemaField('b', 'INTEGER', "
        "'NULLABLE', None, ()),))",
    ]


def test_check_mode_fail_nested():
    local_schema = [
        SchemaField(
            "a", "RECORD", "REQUIRED", fields=[SchemaField("b", "INTEGER", "NULLABLE")]
        )
    ]
    remote_schema = [
        SchemaField(
            "a", "RECORD", "REQUIRED", fields=[SchemaField("b", "INTEGER", "REQUIRED")]
        )
    ]
    assert list(check_schemas(local_schema, remote_schema)) == [
        "Nested: There is difference between SchemaField('b', 'INTEGER', 'NULLABLE', "
        "None, ()) and SchemaField('b', 'INTEGER', 'REQUIRED', None, ())",
    ]


def test_check_field_type_fail():
    local_schema = [SchemaField("a", "STRING", "REQUIRED")]
    remote_schema = [SchemaField("a", "INTEGER", "REQUIRED")]
    assert list(check_schemas(local_schema, remote_schema)) == [
        "There is difference between SchemaField('a', 'STRING', 'REQUIRED', None, ()) "
        "and SchemaField('a', 'INTEGER', 'REQUIRED', None, ())",
    ]


def test_check_field_type_fail_nested():
    local_schema = [
        SchemaField(
            "a", "RECORD", "REQUIRED", fields=[SchemaField("b", "STRING", "REQUIRED")]
        )
    ]
    remote_schema = [
        SchemaField(
            "a", "RECORD", "REQUIRED", fields=[SchemaField("b", "INTEGER", "REQUIRED")]
        )
    ]
    assert list(check_schemas(local_schema, remote_schema)) == [
        "Nested: There is difference between SchemaField('b', 'STRING', 'REQUIRED', "
        "None, ()) and SchemaField('b', 'INTEGER', 'REQUIRED', None, ())",
    ]


def test_check_field_type_unspecified():
    local_schema = [SchemaField("a", "RANDOM_FIELD_TYPE", "REQUIRED")]
    remote_schema = [SchemaField("a", "INTEGER", "REQUIRED")]
    assert list(check_schemas(local_schema, remote_schema)) == [
        "Unspecified field type in SchemaField('a', 'RANDOM_FIELD_TYPE', 'REQUIRED', None, ()) "
        "or SchemaField('a', 'INTEGER', 'REQUIRED', None, ())"
    ]


def test_check_field_type_unspecified_nested():
    local_schema = [
        SchemaField(
            "a",
            "RECORD",
            "REQUIRED",
            fields=[SchemaField("b", "RANDOM_FIELD_TYPE", "REQUIRED")],
        )
    ]
    remote_schema = [
        SchemaField(
            "a", "RECORD", "REQUIRED", fields=[SchemaField("b", "INTEGER", "REQUIRED")]
        )
    ]
    assert list(check_schemas(local_schema, remote_schema)) == [
        "Unspecified field type in SchemaField('b', 'RANDOM_FIELD_TYPE', 'REQUIRED', "
        "None, ()) or SchemaField('b', 'INTEGER', 'REQUIRED', None, ())",
    ]



from datetime import date, datetime, time
from decimal import Decimal
from typing import Dict, NewType, Type

from .big_query_types import BigQueryTypes

Timestamp = NewType("Timestamp", datetime)


TypeMapping: Dict[BigQueryTypes, Type] = {
    BigQueryTypes.STRING: str,
    BigQueryTypes.BYTES: bytes,
    BigQueryTypes.INT64: int,
    BigQueryTypes.FLOAT64: float,
    BigQueryTypes.NUMERIC: Decimal,
    BigQueryTypes.BOOL: bool,
    BigQueryTypes.TIMESTAMP: Timestamp,
    BigQueryTypes.DATE: date,
    BigQueryTypes.TIME: time,
    BigQueryTypes.DATETIME: datetime,
}
